# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import factory
from factory.django import DjangoModelFactory


class UserFactory(DjangoModelFactory):
    class Meta:
        model = 'auth.User'
        inline_args = ('username', 'email', 'password')

    username = factory.Faker('user_name')
    email = factory.Faker('email')
    password = factory.Faker('password')
    is_active = True

    @classmethod
    def _create(cls, model_class, username, email, password, **kwargs):
        instance = model_class.objects._create_user(
            username, email, password, **kwargs)
        instance.raw_password = password
        return instance



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from binascii import unhexlify

from django.test import TestCase
from mock import patch
from yubiotp.otp import YubiKey
from otp_yubikey.models import (
    RemoteYubikeyDevice, ValidationService, default_id)

from kleides_mfa.forms import YubikeyDeviceCreateForm
from kleides_mfa.registry import registry

from .factories import UserFactory


class DjangoOtpYubikeyTestCase(TestCase):
    def login(self, user, redirect_to='/list/'):
        response = self.client.post(
            '/login/',
            {'username': user.username, 'password': user.raw_password},
            follow=True)
        self.assertRedirects(response, redirect_to)
        return response

    @patch.object(RemoteYubikeyDevice, 'verify_token')
    def test_yubikey(self, mock_verify_token):
        user = UserFactory()
        self.login(user)

        # Test bad/missing input.
        mock_verify_token.return_value = False
        service = ValidationService.objects.latest('pk')
        response = self.client.post(
            '/yubikey/create/', {'name': 'Keychain'})
        self.assertContains(
            response, 'This field is required.')
        self.assertTrue(
            response.context['form'].fields['service'].widget.is_hidden)
        # Test field visibility with multiple services.
        ValidationService.objects.create(
            name='YubiCustom', param_sl='', param_timeout='')
        response = self.client.post(
            '/yubikey/create/', {
                'service': service.pk, 'otp_token': 'XXX',
                'name': 'Keychain'})
        self.assertContains(
            response, 'Unable to validate the token with the device.')
        self.assertFalse(
            response.context['form'].fields['service'].widget.is_hidden)

        # Test success.
        mock_verify_token.return_value = True
        yubikey = YubiKey(unhexlify(default_id()), 6, 0)
        registration_token = yubikey.generate()
        response = self.client.post(
            '/yubikey/create/', {
                'service': service.pk, 'otp_token': registration_token,
                'name': 'Keychain'})
        self.assertRedirects(response, '/list/')
        device = user.remoteyubikeydevice_set.get()
        self.assertTrue(device.confirmed)
        self.assertEqual(device.name, 'Keychain')

        # Test update.
        response = self.client.post(
            '/yubikey/update/{}/'.format(device.pk), {'name': 'Acme Inc.'})
        self.assertRedirects(response, '/list/')
        device = user.remoteyubikeydevice_set.get()
        self.assertEqual(device.name, 'Acme Inc.')

        self.client.logout()

        # Login without Yubikey.
        verify_url = '/yubikey/verify/{}/'.format(device.pk)
        response = self.login(user, '{}?next=/list/'.format(verify_url))

        context_user = response.context['user']
        self.assertTrue(context_user.is_anonymous)

        # User is forced to use 2 step authentication.
        mock_verify_token.return_value = False
        response = self.client.get('/list/')
        self.assertRedirects(response, '/login/?next=/list/')
        response = self.client.post('/yubikey/delete/{}/'.format(device.pk))
        self.assertRedirects(
            response, '/login/?next=/yubikey/delete/{}/'.format(device.pk))

        response = self.client.post(verify_url, {'otp_token': ''})
        self.assertContains(
            response, 'This field is required.')

        # Token reuse is not allowed so now the registration token is invalid.
        response = self.client.post(
            verify_url, {'otp_token': registration_token})
        self.assertContains(
            response, 'The token is not valid for this device.')

        # Complete authentication with a new token.
        mock_verify_token.return_value = True
        response = self.client.post(
            verify_url, {'otp_token': yubikey.generate()}, follow=True)
        self.assertRedirects(response, '/list/')

        context_user = response.context['user']
        self.assertTrue(context_user.is_authenticated)
        self.assertTrue(context_user.is_verified)

        # Verify that the authentication method matches.
        self.assertEqual(
            registry.user_authentication_method(context_user), 'yubikey')

        response = self.client.post(
            '/yubikey/delete/{}/'.format(device.pk), follow=True)
        self.assertContains(
            response,
            'The Yubikey &quot;Acme Inc.&quot; was deleted successfully.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.apps import AppConfig, apps
from django.contrib.auth import get_user_model
from django.db.models.signals import post_migrate
from django.utils.translation import ugettext_lazy as _


def is_authenticated(user):
    '''
    User.is_authenticated property that ensures user used 2 step
    authentication.
    '''
    return bool(user.is_verified)


def is_verified(user):
    '''
    User.is_verified property.
    '''
    # Warning: The django-otp version is a function instead of a propertry.
    # otp_device may not be set (django startup check, no middleware).
    return bool(getattr(user, 'otp_device', None) is not None)


def is_single_factor_authenticated(user):
    '''
    User.is_single_factor_authenticated property.
    Always returns True for authenticated users.
    '''
    return bool(not user.is_anonymous)


def create_yubikey_validationservice(app_config, **kwargs):
    '''
    Make sure at least one ValidationService exists.
    '''
    from otp_yubikey.models import ValidationService
    if (app_config.label == 'otp_yubikey' and
            ValidationService.objects.all().count() == 0):
        ValidationService.objects.get_or_create(defaults={
            'name': 'YubiCloud', 'use_ssl': True,
            'param_sl': '', 'param_timeout': ''})


class KleidesMfaConfig(AppConfig):
    name = 'kleides_mfa'
    verbose_name = 'Kleides Multi Factor Authentication'

    def ready(self):
        from . import settings
        from .registry import registry

        if settings.PATCH_USER:
            # Monkey patch user authentication properties.
            User = get_user_model()
            User.is_verified = property(is_verified)
            User.is_authenticated = property(is_authenticated)
            User.is_single_factor_authenticated = property(
                is_single_factor_authenticated)
            from django.contrib.auth.models import AnonymousUser
            AnonymousUser.is_verified = property(is_verified)
            AnonymousUser.is_authenticated = property(is_authenticated)
            AnonymousUser.is_single_factor_authenticated = property(
                is_single_factor_authenticated)

        # Check if known devices are installed and register them as plugins.
        if apps.is_installed('django_otp.plugins.otp_totp'):
            from .forms import TOTPDeviceCreateForm, DeviceVerifyForm
            from django_otp.plugins.otp_totp.models import TOTPDevice
            registry.register(
                'TOTP', TOTPDevice, create_form_class=TOTPDeviceCreateForm,
                verify_form_class=DeviceVerifyForm)

        if apps.is_installed('django_otp.plugins.otp_static'):
            from .forms import RecoveryDeviceForm, DeviceVerifyForm
            from django_otp.plugins.otp_static.models import StaticDevice
            message = _('Your recovery codes have been generated, save them '
                        'somewhere safe! Any old codes you have will no '
                        'longer be usable.')
            delete_message = _('Your recovery codes have been disabled!')
            registry.register(
                'Recovery code', StaticDevice,
                device_list_javascript='js/kleides_mfa/recovery-code.js',
                device_list_template=(
                    'kleides_mfa/device_recovery-code_list.html'),
                create_form_class=RecoveryDeviceForm,
                update_form_class=RecoveryDeviceForm,
                verify_form_class=DeviceVerifyForm,
                create_message=message, update_message=message,
                delete_message=delete_message)

        if apps.is_installed('otp_yubikey'):
            from .forms import DeviceVerifyForm, YubikeyDeviceCreateForm
            from otp_yubikey.models import RemoteYubikeyDevice
            registry.register(
                'Yubikey', RemoteYubikeyDevice,
                create_form_class=YubikeyDeviceCreateForm,
                verify_form_class=DeviceVerifyForm)
            post_migrate.connect(
                create_yubikey_validationservice,
                dispatch_uid='kleides_mfa.apps.KleidesMfaConfig')

        if apps.is_installed('django.contrib.admin') and settings.PATCH_ADMIN:
            from django.contrib import admin
            from .admin import AdminSiteMfaRequiredMixin
            MfaAdminSite = type(
                str('MfaAdminSite'),
                (AdminSiteMfaRequiredMixin, admin.AdminSite), {})
            setattr(admin, 'AdminSite', MfaAdminSite)
            admin.site.__class__ = MfaAdminSite



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import redirect_to_login
from django.shortcuts import resolve_url
from django.utils.http import is_safe_url


class AdminSiteMfaRequiredMixin():
    """
    Mixin for enforcing MFA verified staff users.

    Custom admin views should either be wrapped using :meth:`admin_view` or
    use :meth:`has_permission` in order to secure those views.
    """

    def has_permission(self, request):
        """
        Returns True if the given HttpRequest has permission to view
        *at least one* page in the admin site and uses MFA authentication.
        """
        if not super().has_permission(request):
            return False
        return request.user.is_verified

    def login(self, request, extra_context=None):
        """
        Redirects to the site login page for the given HttpRequest.
        """
        redirect_to = request.POST.get(
            REDIRECT_FIELD_NAME,
            request.GET.get(REDIRECT_FIELD_NAME, '')
        )
        url_is_safe = is_safe_url(
            url=redirect_to,
            allowed_hosts={request.get_host()},
            require_https=request.is_secure(),
        )

        if not redirect_to or not url_is_safe:
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)

        return redirect_to_login(redirect_to)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.contrib.auth import get_user_model, load_backend
from django.contrib.auth.mixins import UserPassesTestMixin
from django.http import Http404
from django.urls import reverse_lazy
from django.utils.crypto import constant_time_compare

from ..registry import registry

# Note that these session keys are different from django auth so the user
# session will never pass as authenticated. Meanwhile we still need to enforce
# the same protections which is the reason for the duplication.
SESSION_KEY = '_kleides-mfa_user_id'
BACKEND_SESSION_KEY = '_kleides-mfa_user_backend'
HASH_SESSION_KEY = '_kleides-mfa_user_hash'
VERIFIED_SESSION_KEY = '_kleides-mfa_user_verified'


class PluginMixin():
    success_url = reverse_lazy('kleides_mfa:index')

    def dispatch(self, *args, **kwargs):
        self.plugin = self.get_plugin()
        return super().dispatch(*args, **kwargs)

    def get_plugin(self):
        try:
            return registry.get_plugin(self.kwargs['plugin'])
        except KeyError:
            raise Http404('Plugin does not exist')

    def get_object(self):
        return self.plugin.get_user_device(
            self.kwargs['device_id'], self.request.user, confirmed=None)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['plugin'] = self.plugin
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['plugin'] = self.plugin
        kwargs['request'] = self.request
        return kwargs

    def get_template_names(self):
        return [
            'kleides_mfa/device_{}{}.html'.format(
                self.plugin.slug, self.template_name_suffix),
            'kleides_mfa/device{}.html'.format(self.template_name_suffix),
        ]


class SingleFactorRequiredMixin(UserPassesTestMixin):
    '''
    Verify that the user is authenticated with a single authentication factor.
    '''
    def test_func(self):
        return self.request.user.is_single_factor_authenticated


class MultiFactorRequiredMixin(UserPassesTestMixin):
    '''
    Verify that the user is authenticated with multiple authentication factors.
    '''
    def test_func(self):
        return self.request.user.is_verified


class SetupOrMFARequiredMixin(UserPassesTestMixin):
    '''
    Verify that the user is authenticated with multiple factors or with single
    factor and is still in the process of account setup.
    '''
    def test_func(self):
        user = self.request.user
        if user.is_verified:
            return True
        return (
            user.is_single_factor_authenticated and
            not registry.user_has_device(user, confirmed=True))


class UnverifiedUserMixin(UserPassesTestMixin):
    '''
    Verify that the session is associated with a User.
    Note that the user may not be fully authenticated.
    '''
    def test_func(self):
        self.unverified_user = self.get_unverified_user()
        return bool(self.unverified_user is not None)

    def get_unverified_user(self):
        '''
        Return the unverified user model instance associated with the session.
        If no user is retrieved return None.
        '''
        user = None
        try:
            user_id = get_user_model()._meta.pk.to_python(
                self.request.session[SESSION_KEY])
            backend_path = self.request.session[BACKEND_SESSION_KEY]
        except KeyError:
            pass
        else:
            if backend_path in settings.AUTHENTICATION_BACKENDS:
                backend = load_backend(backend_path)
                user = backend.get_user(user_id)
                # Verify the session
                if hasattr(user, 'get_session_auth_hash'):
                    session_hash = self.request.session.get(HASH_SESSION_KEY)
                    session_hash_verified = bool(
                        session_hash and constant_time_compare(
                            session_hash,
                            user.get_session_auth_hash()))
                    if not session_hash_verified:
                        self.request.session.flush()
                        user = None

        return user



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Django',
    'django_otp',
]

setup(
    author="Harm Geerts",
    author_email='hgeerts@osso.nl',
    python_requires='>=3.5, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Environment :: Web Environment',
        'Natural Language :: English',
        'Framework :: Django',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        ('Topic :: System :: Systems Administration :: '
         'Authentication/Directory'),
    ],
    description="Interface components to configure and manage multi factor "
                "authentication",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='kleides_mfa',
    name='kleides_mfa',
    packages=find_packages(include=['kleides_mfa', 'kleides_mfa.*']),
    version='0.1.7',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.shortcuts import resolve_url
from django.test import TestCase

from django_otp import DEVICE_ID_SESSION_KEY
from django_otp.plugins.otp_totp.models import TOTPDevice

from kleides_mfa import settings as mfa_settings
from kleides_mfa.forms import DeviceUpdateForm
from kleides_mfa.registry import AlreadyRegistered, registry

from .factories import UserFactory


class KleidesMfaTestCase(TestCase):
    def login(self, user, redirect_to='/list/', login_url='/login/'):
        # The LoginView prepares the session for 2 step authentication.
        response = self.client.post(
            login_url,
            {'username': user.username, 'password': user.raw_password},
            follow=True)
        self.assertRedirects(response, redirect_to)
        return response

    def login_with_mfa(self, user):
        device = user.totpdevice_set.get_or_create(name='test')[0]
        # Force the user to login with 2 step authentication.
        # django-otp should do this if the otp_device is set.
        user.otp_device = device

        # Login may clear the session if a user was logged in.
        self.client.force_login(user)

        # Except the user is taken from the request, the test client does not
        # set it and django login does not add it unless the user attribute
        # already exists...
        session = self.client.session
        session[DEVICE_ID_SESSION_KEY] = device.persistent_id
        session.save()

    def test_login_failure(self):
        response = self.client.post(
            '/login/', {'username': 'test2', 'password': 'test1234'})
        self.assertContains(
            response, 'Please enter a correct username and password.')

    def test_initial_account_setup(self):
        # With single factor authentication a user can see the empty
        # list of available authentication methods as well as add a device.
        # The user cannot do anything else without 2 step authentication.
        user = UserFactory()
        response = self.login(user)
        self.assertContains(
            response, 'These are your 2 step authentication methods.')
        self.assertFalse(registry.user_has_device(user))

        context_user = response.context['user']
        self.assertTrue(context_user.is_single_factor_authenticated)
        self.assertFalse(context_user.is_authenticated)
        self.assertFalse(context_user.is_verified)

        # User is allowed to add a authentication method.
        response = self.client.get('/totp/create/')
        self.assertEqual(response.status_code, 200)

        # Force 2 step login setup.
        self.login_with_mfa(user)
        self.assertTrue(registry.user_has_device(user))

        response = self.client.get('/list/')

        context_user = response.context['user']
        self.assertTrue(context_user.is_single_factor_authenticated)
        self.assertTrue(context_user.is_authenticated)
        self.assertTrue(context_user.is_verified)

        # Single factor authentication no longer authenticates the user.
        self.client.logout()
        device = user.totpdevice_set.get(name='test')
        response = self.login(
            user, '/totp/verify/{}/?next=/list/'.format(device.pk))

        context_user = response.context['user']
        self.assertTrue(context_user.is_anonymous)

        # No access to the device listing with single factor auth.
        response = self.client.get('/list/')
        self.assertRedirects(response, '/login/?next=/list/')
        response = self.client.get('/totp/create/')
        self.assertRedirects(response, '/login/?next=/totp/create/')

        # Start new MFA session.
        self.login_with_mfa(user)

        update_url = '/totp/update/{}/'.format(device.pk)
        response = self.client.get(update_url)
        self.assertEqual(response.status_code, 200)

        context_user = response.context['user']
        self.assertTrue(context_user.is_single_factor_authenticated)
        self.assertTrue(context_user.is_authenticated)
        self.assertTrue(context_user.is_verified)

        # Update TOTP name.
        response = self.client.post(
            '/totp/update/{}/'.format(device.pk), {'name': 'Secure Phone'},
            follow=True)
        self.assertRedirects(response, '/list/')
        self.assertContains(
            response,
            'The TOTP &quot;Secure Phone&quot; was changed successfully.')

        # Secure access is revoked when all authentication methods are removed.
        response = self.client.post(
            '/totp/delete/{}/'.format(device.pk), follow=True)
        self.assertRedirects(response, '/list/')
        self.assertContains(
            response,
            'The TOTP &quot;Secure Phone&quot; was deleted successfully.')

        context_user = response.context['user']
        self.assertTrue(context_user.is_single_factor_authenticated)
        self.assertFalse(context_user.is_authenticated)
        self.assertFalse(context_user.is_verified)

    def test_secure_admin(self):
        # Users must be verified to access the admin interface.
        user = UserFactory(is_staff=True, is_superuser=True)
        self.login(user)

        response = self.client.get('/admin/', follow=True)
        self.assertRedirects(response, '/login/?next=/admin/')

        self.login_with_mfa(user)

        response = self.client.get('/admin/', follow=True)
        self.assertEqual(response.status_code, 200)

    def test_login_redirect(self):
        user = UserFactory()
        # without a device configured the redirect url is SINGLE_FACTOR_URL.
        response = self.login(
            user, redirect_to=resolve_url(mfa_settings.SINGLE_FACTOR_URL))

        device = user.totpdevice_set.create(name='test')
        verify_url = '/totp/verify/{}/?next='.format(device.pk)
        # with a device configured the next redirect url is LOGIN_REDIRECT_URL.
        response = self.login(
            user,
            redirect_to=verify_url + resolve_url(settings.LOGIN_REDIRECT_URL))

        # Unless the user came here with a next parameter.
        response = self.login(
            user, login_url='/login/?next=/some/place/',
            redirect_to=verify_url + '/some/place/')

    def test_plugin(self):
        # Plugins are allowed to use the same device/model.
        totp_plugin = registry.get_plugin('totp')
        registry.register(
            'TEST', TOTPDevice, update_form_class=DeviceUpdateForm)
        plugin = registry.get_plugin('test')
        self.assertEqual(plugin.model, totp_plugin.model)
        self.assertEqual(str(plugin), 'TEST')
        self.assertEqual(
            repr(plugin),
            "KleidesMfaPlugin(name='TEST', model=<class "
            "'django_otp.plugins.otp_totp.models.TOTPDevice'>)")

        self.assertIsNone(plugin.get_create_form_class())
        self.assertIsNotNone(plugin.get_update_form_class())
        self.assertIsNone(plugin.get_verify_form_class())

        # But the name must be unique.
        with self.assertRaises(AlreadyRegistered):
            registry.register('tEsT', TOTPDevice)

        registry.unregister('test')
        with self.assertRaises(KeyError):
            registry.unregister('test')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.test import TestCase

from kleides_mfa.registry import registry

from .factories import UserFactory


class DjangoOtpRecoveryTestCase(TestCase):
    def login(self, user, redirect_to='/list/'):
        response = self.client.post(
            '/login/',
            {'username': user.username, 'password': user.raw_password},
            follow=True)
        self.assertRedirects(response, redirect_to)
        return response

    def test_recovery(self):
        user = UserFactory()
        response = self.login(user)
        self.assertContains(
            response, 'These are your 2 step authentication methods.')

        # Create recovery codes to secure the account.
        response = self.client.post('/recovery-code/create/', follow=True)
        self.assertContains(
            response, 'Your recovery codes have been generated')

        # Login with a recovery code.
        self.client.logout()
        device = user.staticdevice_set.get()
        self.assertEqual(device.token_set.count(), 10)
        device_url = '/recovery-code/verify/{}/'.format(device.pk)
        response = self.login(user, '{}?next=/list/'.format(device_url))

        context_user = response.context['user']
        self.assertTrue(context_user.is_anonymous)

        # Continue authentication using recovery code.
        token = device.token_set.first()
        response = self.client.post(
            device_url, {'otp_token': token.token}, follow=True)
        self.assertRedirects(response, '/list/')
        self.assertEqual(device.token_set.count(), 9)

        context_user = response.context['user']
        self.assertTrue(context_user.is_authenticated)
        self.assertTrue(context_user.is_verified)

        # Verify that the authentication method matches.
        self.assertEqual(
            registry.user_authentication_method(context_user), 'recovery-code')

        # The create and update form for recovery codes replace existing codes.
        self.client.post('/recovery-code/update/{}/'.format(device.pk))
        self.assertEqual(device.token_set.count(), 10)

        # Secure access is revoked when all authentication methods are removed.
        response = self.client.post(
            '/recovery-code/delete/{}/'.format(device.pk), follow=True)
        self.assertRedirects(response, '/list/')

        context_user = response.context['user']
        self.assertTrue(context_user.is_single_factor_authenticated)
        self.assertFalse(context_user.is_authenticated)
        self.assertFalse(context_user.is_verified)
